/*
 * linux/fs/nfsd/nfsfh.c
 *
 * NFS server file handle treatment.
 *
 * Copyright (C) 1995, 1996 Olaf Kirch <okir@monad.swb.de>
 * Portions Copyright (C) 1999 G. Allen Morris III <gam3@acm.org>
 * Extensive rewrite by Neil Brown <neilb@cse.unsw.edu.au> Southern-Spring 1999
 * ... and again Southern-Winter 2000 to support nfsd_operations
 */

#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/dcache.h>
#include <linux/module.h>
#include <asm/pgtable.h>

#include <linux/sunrpc/svc.h>
#include <linux/nfsd/nfsd.h>

#define NFSDDBG_FACILITY		NFSDDBG_FH
#define NFSD_PARANOIA 1
/* #define NFSD_DEBUG_VERBOSE 1 */


static int nfsd_nr_verified;
static int nfsd_nr_put;

extern struct nfsd_operations nfsd_op_default;

#define	CALL(ops,fun) ((ops->fun)?(ops->fun):nfsd_op_default.fun)


struct nfsd_getdents_callback {
	char *name;		/* name that was found. It already points to a buffer NAME_MAX+1 is size */
	unsigned long ino;	/* the inum we are looking for */
	int found;		/* inode matched? */
	int sequence;		/* sequence counter */
};

/*
 * A rather strange filldir function to capture
 * the name matching the specified inode number.
 */
static int filldir_one(void * __buf, const char * name, int len,
			off_t pos, ino_t ino, unsigned int d_type)
{
	struct nfsd_getdents_callback *buf = __buf;
	int result = 0;

	buf->sequence++;
#ifdef NFSD_DEBUG_VERBOSE
dprintk("filldir_one: seq=%d, ino=%ld, name=%s\n", buf->sequence, ino, name);
#endif
	if (buf->ino == ino) {
		memcpy(buf->name, name, len);
		buf->name[len] = '\0';
		buf->found = 1;
		result = -1;
	}
	return result;
}

/**
 * nfsd_get_name - default nfsd_operations->get_name function
 * @dentry: the directory in which to find a name
 * @name:   a pointer to a %NAME_MAX+1 char buffer to store the name
 * @child:  the dentry for the child directory.
 *
 * calls readdir on the parent until it finds an entry with
 * the same inode number as the child, and returns that.
 */
static int nfsd_get_name(struct dentry *dentry, char *name,
		  struct dentry *child)
{
	struct inode *dir = dentry->d_inode;
	int error;
	struct file file;
	struct nfsd_getdents_callback buffer;

	error = -ENOTDIR;
	if (!dir || !S_ISDIR(dir->i_mode))
		goto out;
	error = -EINVAL;
	if (!dir->i_fop)
		goto out;
	/*
	 * Open the directory ...
	 */
	error = init_private_file(&file, dentry, FMODE_READ);
	if (error)
		goto out;
	error = -EINVAL;
	if (!file.f_op->readdir)
		goto out_close;

	buffer.name = name;
	buffer.ino = child->d_inode->i_ino;
	buffer.found = 0;
	buffer.sequence = 0;
	while (1) {
		int old_seq = buffer.sequence;
		error = vfs_readdir(&file, filldir_one, &buffer);
		if (error < 0)
			break;

		error = 0;
		if (buffer.found)
			break;
		error = -ENOENT;
		if (old_seq == buffer.sequence)
			break;
	}

out_close:
	if (file.f_op->release)
		file.f_op->release(dir, &file);
out:
	return error;
}

/**
 * nfsd_get_dentry - default nfsd_operations->get_dentry function
 * sb: 	super_block of target file system.
 * inump: pointer to 32bit inode number followed by 32bit generation number
 *
 * This function abuses iget() to find the inode with a given
 * inode number, and checks that the generation number is correct.
 * It assumes that the filesystems read_inode function will return
 * a "bad_inode" if the inode number is invalid.
 *
 * If the inode is found and it has at least one dentry, the first dentry
 * is returned.
 * If there are no dentrys, one is allocated using d_alloc_root, and
 * it is returned with the DCACHE_NFSD_DISCONNECTED flag set.
 *
 * If a filesystem choses to use this as its get_dentry function, its
 * read_inode() but be able to reliably locate an inode given the inode number,
 * and must also be able to detect an inactive inode, and make the inode structure
 * as bad using make_bad_inode.  Further, the delete_inode() function must be
 * able to detect and ignore a "bad inode".
 *
 * Finally, the filesystem must not depend in have d_op set for a dentry, as
 * this routine may allocate one without setting d_op. 
 */
static struct dentry *nfsd_get_dentry(struct super_block *sb, void *inump)
{
	struct dentry *dentry;
	struct inode *inode;

	__u32 *u32p;
	unsigned long ino;
	__u32 generation;

	u32p = (__u32*)inump;

	ino = *u32p++;
	generation = *u32p;

	if (ino == 0)
		return NULL;
	
	inode = iget(sb, ino);
	if (is_bad_inode(inode)
	    || (generation && inode->i_generation != generation)
		) {
		/* we didn't find the right inode.. */
		dprintk("nfsd_get_dentry: Inode %lu, Bad count: %d %d or version  %u %u\n",
			inode->i_ino,
			inode->i_nlink, atomic_read(&inode->i_count),
			inode->i_generation,
			generation);

		iput(inode);
		return NULL;
	}

	dentry = d_make_alias(inode);
	if (!dentry) {
		iput(inode);
		dentry = ERR_PTR(-ENOMEM);
	}
	return dentry;
}


/**
 * nfsd_find_fh_dentry - helper routine to implement nfsd_operations->decode_fh
 * @sb:		The &super_block identifying the filesystem
 * @obj:	An opaque identifier of the object to be found - passed to get_inode
 * @parent:	An optional opqaue identifier of the parent of the object.
 * @acceptable:	A function used to test possible &dentries to see of they are acceptable
 * @context:	A parameter to @acceptable so that it knows on what basis to judge.
 *
 * nfsd_find_fh_dentry is the central helper routine to enable file systems to provide
 * the decode_fh() nfsd_operation.  It's main task is to take an inode, find or create an
 * appropriate &dentry structure, and possibly splice this into the dcache in the
 * correct place.
 *
 * The decode_fh() operation provided by the filesystem should call nfsd_find_fh_dentry()
 * with the same parameters that it received except that instead of the file handle fragment,
 * pointers to opaque identifiers for the object and optionally its parent are passed.
 * The default decode_fh routine passes one pointer to the start of the filehandle fragment, and
 * one 8 bytes in to the fragment.  It is expected that most filesystems will take this
 * approach, though the offset to the parent identifier may well be different.
 *
 * nfsd_find_fh_dentry() will call get_dentry to get an dentry pointer from the file system.  If
 * any &dentry in the d_alias list is acceptable, it will be returned.  Otherwise
 * nfsd_find_fh_dentry() will attempt to splice a new &dentry into the dcache using get_name() and
 * get_parent() to find the appropriate place.
 *
 */
#ifdef CONFIG_NFSD_MODULE
/* for any other code, nfsd_find_fh_dentry is a macro that
 * dives through nfsd_linkage, but for us, it is a real function
 */

#undef nfsd_find_fh_dentry
#endif

struct dentry *
nfsd_find_fh_dentry(struct super_block *sb, void *obj, void *parent,
		    int (*acceptable)(void *context, struct dentry *de),
		    void *context)
{
	struct dentry *result = NULL;
	struct dentry *target_dir;
	int err;
	struct nfsd_operations *nops = sb->s_nfsd_op;
	struct list_head *le, *head;
	int noprogress;


	/*
	 * Attempt to find the inode.
	 */
	result = CALL(sb->s_nfsd_op,get_dentry)(sb,obj);
	err = -ESTALE;
	if (result == NULL)
		goto err_out;
	if (IS_ERR(result)) {
		err = PTR_ERR(result);
		goto err_out;
	}
	if (S_ISDIR(result->d_inode->i_mode) &&
	    (result->d_flags & DCACHE_NFSD_DISCONNECTED)) {
		/* it is an unconnected directory, we must connect it */
		;
	} else {
		struct dentry *toput = NULL;
		if (acceptable(context, result))
			return result;
		if (S_ISDIR(result->d_inode->i_mode)) {
			/* there is no other dentry, so fail */
			goto err_result;
		}
		/* try any other aliases */
		spin_lock(&dcache_lock);
		head = &result->d_inode->i_dentry;
		list_for_each(le, head) {
			struct dentry *dentry = list_entry(le, struct dentry, d_alias);
			dget_locked(dentry);
			spin_unlock(&dcache_lock);
			if (toput)
				dput(toput);
			toput = NULL;
			if (dentry != result &&
			    acceptable(context, dentry)) {
				dput(result);
				return dentry;
			}
			spin_lock(&dcache_lock);
			toput = dentry;
		}
		spin_unlock(&dcache_lock);
		if (toput)
			dput(toput);
	}			

	/* It's a directory, or we are required to confirm the file's
	 * location in the tree based on the parent information
 	 */
	dprintk("nfs_fh: need to look harder for %d/%d\n",sb->s_dev,*(int*)obj);
	if (S_ISDIR(result->d_inode->i_mode))
		target_dir = dget(result);
	else {
		if (parent == NULL)
			goto err_result;

		target_dir = CALL(sb->s_nfsd_op,get_dentry)(sb,parent);
		if (IS_ERR(target_dir))
			err = PTR_ERR(target_dir);
		if (target_dir == NULL || IS_ERR(target_dir))
			goto err_result;
	}
	/*
	 * Now we need to make sure that target_dir is properly connected.
	 * It may already be, as the flag isn't always updated when connection
	 * happens.
	 * So, we walk up parent links until we find a connected directory,
	 * or we run out of directories.  Then we find the parent, find
	 * the name of the child in that parent, and do a lookup.
	 * This should connect the child into the parent
	 * We then repeat.
	 */

	/* it is possible that a confused file system might no let up complete the
	 * path to the root.  For example, if get_parent returns a directory
	 * in which we cannot find a name for the child.  While this implies a very
	 * sick filesystem we don't want it to cause knfsd to spin.  Hence the noprogress
	 * counter.  If we go through the loop 10 times (2 is probably enough) without
	 * getting anywhere, we just give up
	 */
	noprogress= 0;
	while (target_dir->d_flags & DCACHE_NFSD_DISCONNECTED && noprogress++ < 10) {
		struct dentry *pd = target_dir;
		spin_lock(&dcache_lock);
		while (!IS_ROOT(pd) &&
		       (pd->d_parent->d_flags & DCACHE_NFSD_DISCONNECTED))
			pd = pd->d_parent;

		dget_locked(pd);
		spin_unlock(&dcache_lock);

		if (!IS_ROOT(pd)) {
			/* must have found a connected parent - great */
			pd->d_flags &= ~DCACHE_NFSD_DISCONNECTED;
			noprogress = 0;
		} else if (pd == sb->s_root) {
			printk("nfsd: Eeek filesystem root is not connected, impossible\n");
			pd->d_flags &= ~DCACHE_NFSD_DISCONNECTED;
			noprogress = 0;
		} else {
			/* we have hit the top of a disconnected path.  Try
			 * to find parent and connect
			 * note: racing with some other process renaming a
			 * directory isn't much of a problem here.  If someone
			 * renames the directory, it will end up properly connected,
			 * which is what we want
			 */
			struct dentry *ppd;
			struct dentry *npd;
			char nbuf[NAME_MAX+1];

			down(&pd->d_inode->i_sem);
			ppd = CALL(nops,get_parent)(pd);
			up(&pd->d_inode->i_sem);

			if (IS_ERR(ppd)) {
				err = PTR_ERR(ppd);
				dprintk("nfsfh: get_parent of %ld failed, err %d\n",
					pd->d_inode->i_ino, err);
				dput(pd);
				break;
			}
			dprintk("nfsfh: find name of %lu in %lu\n", pd->d_inode->i_ino, ppd->d_inode->i_ino);
			err = CALL(nops,get_name)(ppd, nbuf, pd);
			if (err) {
				dput(ppd);
				if (err == -ENOENT)
					/* some race between get_parent and get_name?
					 * just try again
					 */
					continue;
				dput(pd);
				break;
			}
			dprintk("nfsfh: found name: %s\n", nbuf);
			down(&ppd->d_inode->i_sem);
			npd = lookup_one(nbuf, ppd);
			up(&ppd->d_inode->i_sem);
			if (IS_ERR(npd)) {
				err = PTR_ERR(npd);
				dprintk("nfsfh: lookup failed: %d\n", err);
				dput(ppd);
				dput(pd);
				break;
			}
			/* we didn't really want npd, we really wanted
			 * a side-effect of the lookup.
			 * hopefully, npd == pd, though it isn't really
			 * a problem if it isn't
			 */
			if (npd == pd)
				noprogress = 0;
			else
				printk("nfsd: npd != pd\n");
			dput(npd);
			dput(ppd);
			if (IS_ROOT(pd)) {
				/* something went wrong, we will have to give up */
				dput(pd);
				break;
			}
		}
		dput(pd);
	}

	if (target_dir->d_flags & DCACHE_NFSD_DISCONNECTED) {
		/* something went wrong - oh-well */
		if (!err)
			err = -ESTALE;
		goto err_target;
	}
	/* if we weren't after a directory, have one more step to go */
	if (result != target_dir) {
		struct dentry *nresult;
		char nbuf[NAME_MAX+1];
		err = CALL(nops,get_name)(target_dir, nbuf, result);
		if (!err) {
			down(&target_dir->d_inode->i_sem);
			nresult = lookup_one(nbuf, target_dir);
			up(&target_dir->d_inode->i_sem);
			if (!IS_ERR(nresult)) {
				if (nresult->d_inode) {
					dput(result);
					result = nresult;
				} else
					dput(nresult);
			}
		}
	}
	dput(target_dir);
	/* now result is properly connected, it is our best bet */
	if (acceptable(context, result))
		return result;
	/* one last try of the aliases.. */
	spin_lock(&dcache_lock);
	head = &result->d_inode->i_dentry;
	list_for_each(le, head) {
		struct dentry *dentry = list_entry(le, struct dentry, d_alias);
		if (dentry != result &&
		    acceptable(context, dentry)) {
			dget_locked(dentry);
			spin_unlock(&dcache_lock);
			dput(result);
			return dentry;
		}
	}
	spin_unlock(&dcache_lock);

	/* drat - I just cannot find anything acceptable */
	dput(result);
	return ERR_PTR(-ESTALE);

 err_target:
	dput(target_dir);
 err_result:
	dput(result);
 err_out:
	if (err == -ESTALE)
		nfsdstats.fh_stale++;
	return ERR_PTR(err);
}


/**
 * nfsd_decode_fh - default nfsd_operations->decode_fh function
 * sb:  The superblock
 * fh:  pointer to the file handle fragment
 * fh_len: length of file handle fragment
 * acceptable: function for testing acceptability of dentrys
 * context:   context for @acceptable
 *
 * This default decode_fh() function assumes that the object identifier
 * is at the start of the fragment, and that the parent identifier, if
 * present, is 8 bytes in.
 */
struct dentry *nfsd_decode_fh(struct super_block *sb, char *fh, int fh_len,
			 int (*acceptable)(void *context, struct dentry *de),
			 void *context)
{
	char *parent = fh+8;
	if (fh_len <=8)
		parent = NULL;
	return nfsd_find_fh_dentry(sb, fh, parent,
				   acceptable, context);
}

/*
 * our acceptability function.
 * if NOSUBTREECHECK, accept anything
 * if not, require that we can walk up to exp->ex_dentry
 * doing some checks on the 'x' bits
 */
int nfsd_acceptable(void *expv, struct dentry *dentry)
{
	struct svc_export *exp = expv;
	struct dentry *tdentry;
	if (exp->ex_flags & NFSEXP_NOSUBTREECHECK)
		return 1;

	for (tdentry = dentry;
	     tdentry != exp->ex_dentry && ! IS_ROOT(tdentry);
	     tdentry = tdentry->d_parent) {
		/* make sure parents give x permission to user */
		if (permission(tdentry->d_parent->d_inode, S_IXOTH)<0)
			break;
	}
	if (tdentry != exp->ex_dentry)
		dprintk("nfsd_acceptable failed at %p %s\n", tdentry, tdentry->d_name.name);
	return tdentry == exp->ex_dentry;
}


/*
 * Perform sanity checks on the dentry in a client's file handle.
 *
 * Note that the file handle dentry may need to be freed even after
 * an error return.
 *
 * This is only called at the start of an nfsproc call, so fhp points to
 * a svc_fh which is all 0 except for the over-the-wire file handle.
 */
u32
fh_verify(struct svc_rqst *rqstp, struct svc_fh *fhp, int type, int access)
{
	struct knfsd_fh	*fh = &fhp->fh_handle;
	struct svc_export *exp;
	struct dentry	*dentry;
	struct inode	*inode;
	u32		error = 0;

	dprintk("nfsd: fh_verify(%s)\n", SVCFH_fmt(fhp));

	if (!fhp->fh_dentry) {
		kdev_t xdev;
		ino_t xino;
		__u32 *datap=NULL;
		int data_left = fh->fh_size/4;
		int nfsdev;
		error = nfserr_stale;
		if (rqstp->rq_vers == 3)
			error = nfserr_badhandle;
		if (fh->fh_version == 1) {
			
			datap = fh->fh_auth;
			if (--data_left<0) goto out;
			switch (fh->fh_auth_type) {
			case 0: break;
			default: goto out;
			}

			switch (fh->fh_fsid_type) {
			case 0:
				if ((data_left-=2)<0) goto out;
				nfsdev = ntohl(*datap++);
				xdev = MKDEV(nfsdev>>16, nfsdev&0xFFFF);
				xino = *datap++;
				break;
			default:
				goto out;
			}
		} else {
			if (fh->fh_size != NFS_FHSIZE)
				goto out;
			/* assume old filehandle format */
			xdev = u32_to_kdev_t(fh->ofh_xdev);
			xino = u32_to_ino_t(fh->ofh_xino);
		}

		/*
		 * Look up the export entry.
		 */
		error = nfserr_stale; 
		exp = exp_get(rqstp->rq_client, xdev, xino);

		if (!exp) {
			/* export entry revoked */
			goto out;
		}

		/* Check if the request originated from a secure port. */
		error = nfserr_perm;
		if (!rqstp->rq_secure && EX_SECURE(exp)) {
			printk(KERN_WARNING
			       "nfsd: request from insecure port (%08x:%d)!\n",
			       ntohl(rqstp->rq_addr.sin_addr.s_addr),
			       ntohs(rqstp->rq_addr.sin_port));
			goto out;
		}

		/* Set user creds if we haven't done so already. */
		nfsd_setuser(rqstp, exp);

		/*
		 * Look up the dentry using the NFS file handle.
		 */

		if (fh->fh_version == 1) {
			if (fh->fh_fileid_type == 0)
				dentry = dget(exp->ex_dentry);
			else {
				struct nfsd_operations *nop = exp->ex_mnt->mnt_sb->s_nfsd_op;
				int len = fh->fh_fileid_type;
				/* compatibility with earlier code.. */
				switch(len) {
				case 1: len = 8; break;
				case 2: len = 12; break;
				}
				if (len > data_left*4) len = data_left*4;
				dentry  = CALL(nop,decode_fh)(exp->ex_mnt->mnt_sb,
							      (char*)datap, len,
							      nfsd_acceptable, exp);
				
				nfsdstats.fh_lookup++;
			}
		} else {
			struct nfsd_operations *nop = exp->ex_mnt->mnt_sb->s_nfsd_op;
			__u32 handle[4];
			handle[0] = fh->ofh_ino;
			handle[1] = fh->ofh_generation;
			handle[2] = fh->ofh_dirino;
			handle[3] = 0;
			dentry = CALL(nop,decode_fh)(exp->ex_mnt->mnt_sb,
						     (char*)handle, fh->ofh_dirino?12:8,
						     nfsd_acceptable, exp);
			nfsdstats.fh_lookup++;
		}

		error = nfserr_stale;
		if (rqstp->rq_vers == 3 && dentry == NULL)
			error = nfserr_badhandle;
		if (dentry == NULL || IS_ERR(dentry))
			goto out;
#ifdef NFSD_PARANOIA
		if (S_ISDIR(dentry->d_inode->i_mode) &&
		    (dentry->d_flags & DCACHE_NFSD_DISCONNECTED)) {
			printk("nfsd: find_fh_dentry returned a DISCONNECTED directory: %s/%s\n",
			       dentry->d_parent->d_name.name, dentry->d_name.name);
		}
#endif

		fhp->fh_dentry = dentry;
		fhp->fh_export = exp;
		nfsd_nr_verified++;
	} else {
		/* just rechecking permissions
		 * (e.g. nfsproc_create calls fh_verify, then nfsd_create does as well)
		 */
		dprintk("nfsd: fh_verify - just checking\n");
		dentry = fhp->fh_dentry;
		exp = fhp->fh_export;
	}

	inode = dentry->d_inode;

	/* Type check. The correct error return for type mismatches
	 * does not seem to be generally agreed upon. SunOS seems to
	 * use EISDIR if file isn't S_IFREG; a comment in the NFSv3
	 * spec says this is incorrect (implementation notes for the
	 * write call).
	 */

	/* Type can be negative when creating hardlinks - not to a dir */
	if (type > 0 && (inode->i_mode & S_IFMT) != type) {
		error = (type == S_IFDIR)? nfserr_notdir : nfserr_isdir;
		goto out;
	}
	if (type < 0 && (inode->i_mode & S_IFMT) == -type) {
		error = (type == -S_IFDIR)? nfserr_notdir : nfserr_isdir;
		goto out;
	}

	/* Finally, check access permissions. */
	error = nfsd_permission(exp, dentry, access);
#ifdef NFSD_PARANOIA_EXTREME
	if (error) {
		printk("fh_verify: %s/%s permission failure, acc=%x, error=%d\n",
		       dentry->d_parent->d_name.name, dentry->d_name.name, access, (error >> 24));
	}
#endif
out:
	if (error == nfserr_stale)
		nfsdstats.fh_stale++;
	return error;
}



/**
 * nfsd_encode_fh - default nfsd_operations->encode_fh function
 * dentry:  the dentry to encode
 * fh:      where to stor the file handle fragment
 * max_len: maximum length to store there
 * connectable: whether to store parent infomation
 *
 * This default encode_fh function assumes that the 32 inode number
 * is suitable for locating an inode, and that the generation number
 * can be used to check that it is still valid.  It places them in the
 * filehandle fragment where nfsd_decode_fh expects to find them.
 */
int nfsd_encode_fh(struct dentry *dentry, char *fh, int max_len,
		   int connectable)
{
	struct inode * inode = dentry->d_inode;
	struct inode *parent = dentry->d_parent->d_inode;
	__u32 new[4];
	int cnt = 8;
	
	if (max_len < 8 || (connectable && max_len < 16))
		return -ENOSPC;

	new[0] = inode->i_ino;
	new[1] = inode->i_generation;
	if (connectable && !S_ISDIR(inode->i_mode)) {
		new[2] = parent->i_ino;
		new[3] = parent->i_generation;
		cnt= 16;
	}
	memcpy(fh, new, cnt);
	return cnt;
}



/*
 * Compose a file handle for an NFS reply.
 *
 * Note that when first composed, the dentry may not yet have
 * an inode.  In this case a call to fh_update should be made
 * before the fh goes out on the wire ...
 */
inline int _fh_update(struct dentry *dentry, struct svc_export *exp,
		      __u32 **datapp, int maxsize)
{
	struct nfsd_operations *nop = exp->ex_mnt->mnt_sb->s_nfsd_op;
	int len, len2;
	char *datap = (char*) *datapp;
	
	if (dentry == exp->ex_dentry)
		return 0;

	len = CALL(nop,encode_fh)(dentry, datap, maxsize,
			     !(exp->ex_flags&NFSEXP_NOSUBTREECHECK));
	if (len<0)
		return len;

	/* round to four-byte boundry */
	len2=len;
	while (len2&3)
		datap[len2++] = 0;
	*datapp = (__u32*) (datap+len2);
	return len;
}

int
fh_compose(struct svc_fh *fhp, struct svc_export *exp, struct dentry *dentry)
{
	struct inode * inode = dentry->d_inode;
	struct dentry *parent = dentry->d_parent;
	__u32 *datap;
	int err;

	dprintk("nfsd: fh_compose(exp %x/%ld %s/%s, ino=%ld)\n",
		exp->ex_dev, (long) exp->ex_ino,
		parent->d_name.name, dentry->d_name.name,
		(inode ? inode->i_ino : 0));

	if (fhp->fh_locked || fhp->fh_dentry) {
		printk(KERN_ERR "fh_compose: fh %s/%s not initialized!\n",
			parent->d_name.name, dentry->d_name.name);
	}
	if (fhp->fh_maxsize < NFS_FHSIZE)
		printk(KERN_ERR "fh_compose: called with maxsize %d! %s/%s\n",
		       fhp->fh_maxsize, parent->d_name.name, dentry->d_name.name);

	fhp->fh_dentry = dentry; /* our internal copy */
	fhp->fh_export = exp;

	fhp->fh_handle.fh_version = 1;
	fhp->fh_handle.fh_auth_type = 0;
	fhp->fh_handle.fh_fsid_type = 0;
	datap = fhp->fh_handle.fh_auth+0;
	/* fsid_type 0 == 2byte major, 2byte minor, 4byte inode */
	*datap++ = htonl((MAJOR(exp->ex_dev)<<16)| MINOR(exp->ex_dev));
	*datap++ = ino_t_to_u32(exp->ex_ino);

	if (inode) {
		err = _fh_update(dentry, exp, &datap, fhp->fh_maxsize-3*4);
		if (err < 0)
			return nfserr_opnotsupp;
		fhp->fh_handle.fh_fileid_type = err;
	}

	fhp->fh_handle.fh_size = (datap-fhp->fh_handle.fh_auth+1)*4;

	nfsd_nr_verified++;
	return 0;
}

/*
 * Update file handle information after changing a dentry.
 * This is only called by nfsd_create, nfsd_create_v3 and nfsd_proc_create
 */
int
fh_update(struct svc_fh *fhp)
{
	struct dentry *dentry;
	__u32 *datap;
	int err;
	
	if (!fhp->fh_dentry)
		goto out_bad;

	dentry = fhp->fh_dentry;
	if (!dentry->d_inode)
		goto out_negative;
	if (fhp->fh_handle.fh_fileid_type != 0)
		goto out_uptodate;
	datap = fhp->fh_handle.fh_auth+
		          fhp->fh_handle.fh_size/4 -1;
	err =_fh_update(dentry, fhp->fh_export, &datap, fhp->fh_maxsize-fhp->fh_handle.fh_size);
	if (err < 0)
		return nfserr_opnotsupp;
	fhp->fh_handle.fh_fileid_type = err;
	fhp->fh_handle.fh_size = (datap-fhp->fh_handle.fh_auth+1)*4;
out:
	return 0;

out_bad:
	printk(KERN_ERR "fh_update: fh not verified!\n");
	goto out;
out_negative:
	printk(KERN_ERR "fh_update: %s/%s still negative!\n",
		dentry->d_parent->d_name.name, dentry->d_name.name);
	goto out;
out_uptodate:
	printk(KERN_ERR "fh_update: %s/%s already up-to-date!\n",
		dentry->d_parent->d_name.name, dentry->d_name.name);
	goto out;
}

/*
 * Release a file handle.
 */
void
fh_put(struct svc_fh *fhp)
{
	struct dentry * dentry = fhp->fh_dentry;
	if (dentry) {
		fh_unlock(fhp);
		fhp->fh_dentry = NULL;
		dput(dentry);
		nfsd_nr_put++;
	}
	return;
}

static struct dentry *nfsd_get_parent(struct dentry *child)
{
	/* get_parent cannot be supported generically, the locking
	 * is too icky.
	 * instead, we just return EACCES.  If server reboots or inodes
	 * get flused, you lose
	 */
	return ERR_PTR(-EACCES);
}


struct nfsd_operations nfsd_op_default = {
	decode_fh:	nfsd_decode_fh,
	encode_fh:	nfsd_encode_fh,

	get_name:	nfsd_get_name,
	get_parent:	nfsd_get_parent,
	get_dentry:	nfsd_get_dentry,
};

#ifndef CONFIG_NFSD_MODULE
/* we don't export this when compiling as a module, as
 * we use the nfsd_linkage structure for linkage instead
 */
EXPORT_SYMBOL(nfsd_find_fh_dentry);
#endif
