/*
 *  linux/fs/supermount/dir.c
 *
 *  Original version:
 *      Copyright (C) 1995, 1997
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/dir.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/dir.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *
 *  Rewriten for kernel 2.2 & 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@abc.cap.ru)
 *
 *  $Id: dir.c,v 1.43 2000/10/31 16:17:18 root Exp $
 */

#include <asm/segment.h>
#include <asm/current.h>

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/supermount_fs.h>
#include <linux/stat.h>
#include <linux/sched.h>
#include <linux/file.h>
#include <linux/poll.h>
#include <linux/locks.h>
#include <linux/malloc.h>
#include <asm/smplock.h>

#define DEFAULT_POLLMASK (POLLIN | POLLOUT | POLLRDNORM | POLLWRNORM)

static int supermount_open (struct inode *, struct file *);
static loff_t supermount_lseek (struct file *, loff_t, int);
static ssize_t supermount_read (struct file *, char *, size_t, loff_t *);
static int supermount_readdir (struct file *, void *, filldir_t);
static int supermount_ioctl (struct inode *, struct file *, unsigned int, 
				 unsigned long);
static int supermount_flush (struct file *);
static int supermount_fsync (struct file *, struct dentry *, int datasync);
static int supermount_release (struct inode *, struct file *);
static int supermount_fasync (int, struct file *, int);
static int supermount_setattr (struct dentry *dentry, struct iattr *attr);
static ssize_t supermount_write (struct file *file, const char *buf,
				 size_t count, loff_t *ppos);
static unsigned int supermount_poll (struct file *file,
				     struct poll_table_struct *table);
static int supermount_lock (struct file *file, int cmd, 
			    struct file_lock *fl);

struct file_operations supermount_dir_operations = {
	0,
	supermount_lseek,	/* lseek */
	supermount_read,	/* read */
	NULL,			/* write - bad */
	supermount_readdir,	/* readdir */
	NULL,			/* poll */
	supermount_ioctl,	/* ioctl */
	NULL,			/* mmap */
	supermount_open,	/* open */
	supermount_flush,	/* flush */
	supermount_release,	/* release */
	supermount_fsync,	/* fsync */
	supermount_fasync,	/* fasync */
	NULL,			/* lock */
	NULL,			/* readv */
	NULL			/* writev */
};

struct file_operations supermount_file_operations = {
	0,
	supermount_lseek,		/* lseek */
	supermount_read,		/* read */
	supermount_write,		/* write */
	NULL,				/* readdir */
	supermount_poll,		/* poll */
	supermount_ioctl,		/* ioctl */
	generic_file_mmap,		/* mmap */
	supermount_open,		/* open */
	supermount_flush,		/* flush */
	supermount_release,		/* release */
	supermount_fsync,		/* fsync */
	supermount_fasync,		/* fasync */
	supermount_lock,		/* lock */
	NULL,				/* readv */
	NULL				/* writev */
};

/*
 * directories can handle most operations...  supermount/namei.c just
 * passes them through to the underlying subfs, except for lookup().
 */
struct inode_operations supermount_dir_iops = {
//	&supermount_dir_operations,	/* default directory file-ops */
	supermount_create,		/* create */
	supermount_lookup,		/* lookup */
	NULL/*supermount_link*/,	/* link */
	supermount_unlink,		/* unlink */
	supermount_symlink,		/* symlink */
	supermount_mkdir,		/* mkdir */
	supermount_rmdir,		/* rmdir */
	supermount_mknod,		/* mknod */
	supermount_rename,		/* rename */
	NULL,				/* readlink */
	NULL,				/* follow_link */
	NULL,				/* truncate */
	supermount_permission,		/* permission */
	NULL,				/* revalidate */
	supermount_setattr,		/* setattr */
	NULL				/* getattr */
};

struct inode_operations supermount_file_iops = {
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* truncate */
	NULL,			/* permission */
	NULL,			/* revalidate */
	supermount_setattr,	/* setattr */
	NULL			/* getattr */
};

static struct dentry *__get_subfs_dentry(struct dentry *dentry, 
					 struct dentry *super_root)
{
	struct dentry *parent_subdent;
	struct dentry *subdent = NULL;
	char *tmp;
	char *path;
	struct nameidata nd;
	struct list_head *p;
	struct vfsmount *tmp_m;

	if (dentry == NULL)
	{
		supermount_debug("No dentry\n");
		return NULL;
	}

	tmp = __getname();
	if (tmp == NULL)
	{
		supermount_debug("No page\n");
		return NULL;
	}
	path = tmp + PATH_MAX;
	*path = 0;
		
	supermount_debug("kl: %d\n", kernel_locked());
	spin_lock(&dcache_lock);
	for (;;)
	{
		supermount_debug("%s %d %08x:%08x\n", dentry->d_name.name, PATH_MAX, dentry, dentry->d_parent);
		if (dentry->d_parent == dentry)
		{
			if (*path != '/')
			{
				if (--path < tmp)
					goto err_out;
				*path = '/';
			}
			
			break;
		}
		path -= dentry->d_name.len + 1;
		if (path < tmp)
			goto err_out;
		memcpy(path + 1, dentry->d_name.name, dentry->d_name.len);
		*path = '/';
		dentry = dentry->d_parent;
	}
	supermount_debug("%s\n", path);

	nd.last_type = LAST_ROOT;
	nd.flags = LOOKUP_FOLLOW;

	if (!super_root)
	{
		p = dentry->d_sb->u.supermount_sb.s_undermount->d_vfsmnt.next;
		if (p == &dentry->d_sb->u.supermount_sb.s_undermount->d_vfsmnt)
			goto err_out;
		nd.mnt = mntget(list_entry(p, struct vfsmount, mnt_clash));
		nd.dentry = dget(dentry->d_sb->u.supermount_sb.s_subfs->s_root);
	}
	else
	{
		nd.mnt = 0;
		p = super_root->d_sb->s_mounts.next;
	    	while (p != &super_root->d_sb->s_mounts)
		{
			tmp_m = list_entry(p, struct vfsmount, mnt_instances);
			if (tmp_m->mnt_sb == super_root->d_sb)
			{
				nd.mnt = mntget(tmp_m->mnt_parent);
				break;
			}
			p = p->next;
		}
		if (nd.mnt != 0)
			nd.dentry = dget(super_root);
	}
	spin_unlock(&dcache_lock);
	if (nd.mnt == 0 || path_walk(path, &nd))
		goto out;
	mntput(nd.mnt);
	subdent = nd.dentry;
	goto out;

err_out:
	spin_unlock(&dcache_lock);
out:
	putname(tmp);

	return subdent;
}

struct dentry *get_subfs_dentry(struct dentry *dentry)
{
	struct dentry *subdent = 0;
	
	if (supermount_media_check(dentry->d_sb)) {
		supermount_debug("media is invalid\n");
		return 0;
	}

	if (!subfs_is_active(dentry->d_sb)) {
		supermount_debug ("going online.\n");
		dentry->d_sb->u.supermount_sb.s_state = 
			SUPERMOUNT_ONLINE;
	}

	subdent = __get_subfs_dentry(dentry, 0);

	return subdent;
}

struct dentry *get_subfs_dentry_by_inode(struct inode *inode)
{
	struct dentry *ret = NULL;
	struct dentry *alias = d_find_alias(inode);

	if (inode->i_ino == SUPERMOUNT_ROOT_INO)
		ret = get_subfs_dentry(inode->i_sb->s_root);
	else if (alias)
		ret = get_subfs_dentry(alias);
	if (alias)
		dput(alias);
	
	return ret;
}

struct dentry *get_superfs_dentry_by_inode(struct inode *inode)
{
	struct dentry *ret = NULL;
	struct dentry *alias = d_find_alias(inode);
	
	supermount_debug("asd\n");

	if (alias && inode->i_shadow)
	{
		if (supermount_media_check(inode->i_shadow)) {
			supermount_debug("media is invalid\n");
			return 0;
		}

		if (!subfs_is_active(inode->i_shadow)) {
			supermount_debug ("going online.\n");
			inode->i_shadow->u.supermount_sb.s_state = 
				SUPERMOUNT_ONLINE;
		}

		if (alias && inode->i_shadow)
		{
			supermount_debug("doing actual lookup\n");

			ret = __get_subfs_dentry(alias, inode->i_shadow->s_root);
		}
	}
	if (alias)
		dput(alias);

	supermount_debug("result: %08x\n", ret);
	
	return ret;
}

void wait_on_mount(struct super_block * sb);

/**** on chardev opening kernel is locked ****/

int supermount_open (struct inode * inode, struct file * file)
{
	struct dentry *subdent;
	struct file * subfile;
	int rc = -ENFILE;

	supermount_debug("pwd: %08lx, %08lx, %d, %s\n",
		(unsigned long)current->fs->pwd,
		(unsigned long)current->fs->pwd->d_parent,
		list_empty(&current->fs->pwd->d_hash),
		current->fs->pwd->d_name.name);
	supermount_debug("open inode %ld, name: %s\n", inode->i_ino, file->f_dentry->d_name.name);
	supermount_debug("File dentry count: %d\n",
		atomic_read(&file->f_dentry->d_count));

	if (S_ISDIR(inode->i_mode) && (file->f_mode & FMODE_WRITE))
		return -EPERM;

	subfile = get_empty_filp();
	if (!subfile)
		goto ret;

	rc = -EINVAL;
	subdent = get_subfs_dentry(file->f_dentry);
	if (subdent == NULL)
		goto put_filp_ret;
	if (subdent->d_inode == NULL)
		goto dput_ret;

	supermount_debug("%d\n", atomic_read(&subdent->d_count));
	supermount_iopen(inode->i_sb);

	if (file->f_mode & FMODE_WRITE)
	{
		rc = supermount_get_write_access (inode->i_sb);
		if (rc)
			goto close_ret;
	}

	file->private_data = (void *) subfile;
	
	subfile->f_dentry = subdent;
	if (subdent->d_inode)
		subfile->f_op = subdent->d_inode->i_fop;
	subfile->f_pos = 0;
	subfile->f_reada = 0;
	subfile->f_flags = file->f_flags;
	subfile->f_mode = file->f_mode;

	rc = 0;
	if (subfile->f_op && subfile->f_op->open) {
		rc = subfile->f_op->open(subdent->d_inode, subfile);
		if (rc) {
			subfile->f_dentry = NULL;
			subfile->f_vfsmnt = NULL;
			goto close_ret;
		}
	}
	supermount_debug("%d\n", atomic_read(&subdent->d_count));

	return rc;

close_ret:
	supermount_iclose(inode->i_sb);
dput_ret:
	dput(subdent);
put_filp_ret:
	put_filp(subfile);
	supermount_go_inactive(inode->i_sb);
ret:
	return rc;
}

/*
 * Pass on an lseek() to the underlying subfs
 */

static loff_t supermount_lseek (struct file *file, loff_t offset, int origin)
{
	struct file *subfile;
	struct inode *subinode;
	loff_t retval;

	supermount_debug ("dir_lseek inode %ld\n", file->f_dentry->d_inode->i_ino);

	if ((subfile = (struct file *) file->private_data) == 0)
	{
		retval = 0;
		goto out;
	}
	if ((subinode = subfile->f_dentry->d_inode) == 0) {
		supermount_panic (file->f_dentry->d_inode->i_sb, "supermount_dir_lseek", 
				  "No hidden inode on open directory");
	}
	
	retval = do_lseek(subfile, offset, origin);
	file->f_pos = subfile->f_pos;
	file->f_reada = subfile->f_reada;
	file->f_version = subfile->f_version;

out:
	return retval;
}

static ssize_t supermount_read (struct file *file, char *buf, size_t count,
				loff_t *ppos)
{
	struct file *subfile;
	struct inode *subinode;
	int rc = -EINVAL;

	if ((subfile = (struct file *) file->private_data) == 0)
	{
		rc = 0;
		goto out;
	}
	if ((subinode = subfile->f_dentry->d_inode) == 0) {
		supermount_panic (file->f_dentry->d_inode->i_sb, "supermount_read", 
				  "No hidden inode on open directory");
	}

	supermount_debug ("read inode %ld, subinode: %ld/%ld\n",
		file->f_dentry->d_inode->i_ino, subinode->i_ino,
		subfile->f_dentry->d_inode->i_ino);

	if (subfile->f_op && subfile->f_op->read)
		rc = subfile->f_op->read(subfile, buf, count, ppos);
	if (rc > 0)
		file->f_pos = subfile->f_pos = *ppos;

	supermount_dump_normal_inode(subinode);

out:
	return rc;
}

struct inode *get_possible_super_inode(struct super_block *sb, ino_t ino);

/*
 * All magic in this function is because of obsolete inodes still present in
 * filesystem
 */
int supermount_readdir (struct file *file, void * buf, filldir_t fill_fn)
{
	struct file *subfile;
	int rc = -EINVAL;

	int supermount_filldir_callback(void * __buf, const char * name, 
					int namlen, off_t offset, ino_t ino, 
					unsigned int d_type)
	{
		struct qstr qname = {name, namlen, full_name_hash(name, namlen)};
		ino_t tmp_ino = find_inode_number(file->f_dentry, &qname);
		struct inode *tmp;
/*		supermount_debug("%d, %*s\n", namlen, namlen, name);*/
		if (tmp_ino)
		{
			supermount_debug("found superinode: %ld/%ld/%ld\n",
				file->f_dentry->d_inode->i_ino, ino, tmp_ino);
			ino = tmp_ino;
		}
		else
		{
			tmp = get_possible_super_inode(file->f_dentry->d_sb, ino);
			supermount_debug("new superinode: %ld/%ld/%ld\n",
				file->f_dentry->d_inode->i_ino, ino, tmp->i_ino);
			ino = tmp->i_ino;
			iput(tmp);
		}
		return fill_fn(__buf, name, namlen, offset, ino, d_type);
	}

	if (!S_ISDIR(file->f_dentry->d_inode->i_mode))
		return -EINVAL;
	
	supermount_debug ("readdir inode %ld\n", file->f_dentry->d_inode->i_ino);

	if ((subfile = (struct file *) file->private_data) == 0)
	{
		rc = 0;
		goto out;
	}
	
	if (subfile->f_op && subfile->f_op->readdir)
		rc = subfile->f_op->readdir(subfile, buf, supermount_filldir_callback);
	file->f_pos = subfile->f_pos; /* update position, nfs needs it */

out:
	return rc;
}

static int supermount_ioctl (struct inode *inode, struct file *file,
			     unsigned int cmd, unsigned long arg)
{
	struct file *subfile;
	struct inode *subinode;
	int rc = -ENOTTY;

	supermount_debug ("dir_ioctl inode %ld\n", inode->i_ino);

	if ((subfile = (struct file *) file->private_data) == 0)
		goto out;
	if ((subinode = subfile->f_dentry->d_inode) == 0) {
		supermount_panic (file->f_dentry->d_inode->i_sb, "supermount_ioctl", 
				  "No hidden inode on open directory");
	}

	if (subfile->f_op && subfile->f_op->ioctl)
		rc = subfile->f_op->ioctl(subinode, subfile, cmd, arg);

out:
	return rc;
}

static int supermount_flush (struct file *file)
{
	struct file *subfile;
	int rc = 0;

	supermount_dump_inode(file->f_dentry->d_inode);
	supermount_debug("kl: %d\n", kernel_locked());

	if ((subfile = (struct file *) file->private_data) == 0)
		goto out;

	if (subfile->f_op && subfile->f_op->flush)
		rc = subfile->f_op->flush(subfile);

out:
	return 0;
}

static int supermount_fsync (struct file *file, struct dentry *dentry, int datasync)
{
	struct file *subfile;
	int rc = -EINVAL;

	supermount_dump_inode(file->f_dentry->d_inode);

	if ((subfile = (struct file *) file->private_data) == 0)
	{
		rc = 0;
		goto out;
	}

	if (subfile->f_op && subfile->f_op->fsync)
		rc = subfile->f_op->fsync(subfile, subfile->f_dentry, datasync);

out:
	return rc;
}

static int supermount_fasync (int fd, struct file *file, int on)
{
	struct file *subfile;
	int rc = -EINVAL;

	supermount_dump_inode(file->f_dentry->d_inode);

	if ((subfile = (struct file *) file->private_data) == 0)
	{
		rc = 0;
		goto out;
	}

	if (subfile->f_op && subfile->f_op->fasync)
		rc = subfile->f_op->fasync(fd, subfile, on);

out:
	return -EINVAL;
}

static int supermount_release (struct inode *inode, struct file *file)
{
	struct file *subfile;
	struct inode *subinode;
//	struct dentry *d = subfile->f_dentry;
	int write_on = 0, c;

	supermount_debug ("release inode %ld\n", inode->i_ino);
	supermount_debug("kl: %d\n", kernel_locked());

	supermount_debug("Root count: %d\n", atomic_read(&inode->i_sb->s_root->d_count));
	supermount_debug("Subroot count: %d\n",
		atomic_read(&inode->i_sb->u.supermount_sb.s_subfs->s_root->d_count));
	supermount_dump_normal_inode(inode->i_sb->u.supermount_sb.s_subfs->s_root->d_inode);

	if ((subfile = (struct file *) file->private_data) == 0)
		goto out;

	supermount_debug("Subfile dentry: %08lx\n", (unsigned long)subfile->f_dentry);
	if ((subinode = subfile->f_dentry->d_inode) == 0) {
		supermount_panic (file->f_dentry->d_inode->i_sb, "supermount_dir_lseek", 
				  "No hidden inode on open directory");
	}
	if (!(file->f_mode & FMODE_WRITE) && subinode->i_nlink == 0)
	{
		write_on = 1;
		supermount_get_write_access (inode->i_sb);
	}

	c = atomic_read(&subfile->f_dentry->d_count);
	filp_close (subfile, current->files);
//	supermount_debug("%d\n", atomic_read(&d->d_count));
	if ((file->f_mode & FMODE_WRITE) || write_on)
	{
		mark_subfs_dirty(inode->i_sb);
		supermount_put_write_access (inode->i_sb);
	}
	supermount_iclose(inode->i_sb);

	supermount_debug("Subroot count: %d\n",
		atomic_read(&inode->i_sb->u.supermount_sb.s_subfs->s_root->d_count));
	supermount_dump_normal_inode(inode->i_sb->u.supermount_sb.s_subfs->s_root->d_inode);
	supermount_debug("Subfile dentry: %08lx\n", (unsigned long)subfile->f_dentry);

	supermount_dump_inode(inode);

	supermount_debug("File dentry count: %d/%d, list: %d/%d\n",
		atomic_read(&file->f_dentry->d_count), c,
		list_empty(&file->f_dentry->d_lru),
		list_empty(&file->f_dentry->d_hash)
		);

out:
	return 0;
}

static int supermount_setattr (struct dentry *dentry, struct iattr *attr)
{
	struct inode *subi, *inode = dentry->d_inode;
	struct dentry *subdent;
	int rc = -ENOENT;

	if (!inode)
		return -EINVAL;

	supermount_dump_inode (inode);

	subdent = get_subfs_dentry(dentry);
	if (!subdent)
		goto ret;
	if (!(subi = subdent->d_inode))
		goto dput_ret;

	supermount_iopen(inode->i_sb);

	rc = supermount_get_write_access(inode->i_sb);
	if (rc)
		goto close_ret;

	if (subi->i_op && subi->i_op->setattr)
		rc = subi->i_op->setattr(dentry, attr);
	else {
		rc = inode_change_ok(subi, attr);
		if (!rc)
			inode_setattr(subi, attr);
	}

	supermount_debug("rc: %d\n", rc);

	if (!rc) {
		/* 
		 * If it worked, then we need to mark the modification
		 * to the subfs, and we also need to propogate the
		 * change up to the shadowing inode.  
		 */
		mark_subfs_dirty(inode->i_sb);
		attr->ia_mode = subi->i_mode;
		attr->ia_uid = subi->i_uid;
		attr->ia_gid = subi->i_gid;
		attr->ia_size = subi->i_size;
		attr->ia_atime = subi->i_atime;
		attr->ia_mtime = subi->i_mtime;
		attr->ia_ctime = subi->i_ctime;
		attr->ia_valid = ATTR_UID | ATTR_GID | ATTR_MODE | ATTR_SIZE | 
						 ATTR_ATIME | ATTR_MTIME | ATTR_CTIME;
		supermount_debug("mode: %06o\n", attr->ia_mode);
		inode_setattr(inode, attr);
		supermount_debug("mode: %06o\n", inode->i_mode);
	}

	supermount_put_write_access(inode->i_sb);
close_ret:
	supermount_iclose(inode->i_sb);
dput_ret:
	dput(subdent);
ret:
	supermount_go_inactive(inode->i_sb);

	return rc;
}

static ssize_t supermount_write (struct file *file, const char *buf,
				 size_t count, loff_t *ppos)
{
	struct file *subfile;
	struct inode *subinode;
	int rc = -EINVAL;

	if ((subfile = (struct file *) file->private_data) == 0)
	{
		rc = 0;
		goto out;
	}
	if ((subinode = subfile->f_dentry->d_inode) == 0) {
		supermount_panic (file->f_dentry->d_inode->i_sb, "supermount_dir_lseek", 
				  "No hidden inode on open directory");
	}
	supermount_debug ("write inode %ld, subinode: %ld/%ld\n",
		file->f_dentry->d_inode->i_ino, subinode->i_ino,
		subfile->f_dentry->d_inode->i_ino);
	supermount_debug("%Ld, %d, %Ld\n", subinode->i_size, count, *ppos);

	if (subfile->f_op && subfile->f_op->write)
		rc = subfile->f_op->write(subfile, buf, count, ppos);
	supermount_debug("rc: %d\n", rc);
	if (rc > 0)
	{
		supermount_debug("%Ld, %d, %Ld, %Ld\n", subinode->i_size, count, 
			*ppos, subfile->f_pos);
		file->f_pos = subfile->f_pos = *ppos;
		file->f_mode = subfile->f_mode;
		file->f_dentry->d_inode->i_size = subinode->i_size;
		file->f_dentry->d_inode->i_blocks = subinode->i_blocks;
		file->f_dentry->d_inode->i_mode = subinode->i_mode;
		supermount_debug("%Ld, %d, %Ld\n", subinode->i_size, count, *ppos);
	}

out:
	return rc;
}

static unsigned int supermount_poll (struct file *file,
				     struct poll_table_struct *table)
{
	struct file *subfile;
	struct inode *subinode;
	int rc = DEFAULT_POLLMASK;
	
	if ((subfile = (struct file *) file->private_data) == 0)
		goto out;
	if ((subinode = subfile->f_dentry->d_inode) == 0) {
		supermount_panic (file->f_dentry->d_inode->i_sb, "supermount_dir_lseek", 
				  "No hidden inode on open directory");
	}
	supermount_debug ("poll inode %ld, subinode: %ld/%ld\n",
		file->f_dentry->d_inode->i_ino, subinode->i_ino,
		subfile->f_dentry->d_inode->i_ino);


	if (subfile->f_op && subfile->f_op->poll)
		rc = subfile->f_op->poll(subfile, table);

out:
	return rc;
}

static int supermount_lock (struct file *file, int cmd, 
			    struct file_lock *fl)
{
	struct file *subfile;
	int rc = 0;

	supermount_dump_inode(file->f_dentry->d_inode);

	if ((subfile = (struct file *) file->private_data) == 0)
		goto out;

	if (subfile->f_op && subfile->f_op->lock)
		rc = subfile->f_op->lock(subfile, cmd, fl);
	else
		if (cmd == F_GETLK)
			posix_test_lock(file, fl);

out:
	return rc;
}

static int dir_dentry_revalidate(struct dentry *dentry, int flags)
{
	struct dentry *subd;
	int rc = 1;
	
	supermount_debug("revalidate dir dentry %s\n", dentry->d_name.name);
	if (!supermount_revalidate(dentry->d_sb))
	{
		d_drop(dentry);
		return 0;
	}

	if ((subd = get_subfs_dentry(dentry)) == 0)
	{
		rc = 0;
		d_drop(dentry);
		goto ret;
	}
	if (subd->d_op && subd->d_op->d_revalidate)
	{
		supermount_debug("lowlevel\n");
		rc = subd->d_op->d_revalidate(subd, flags);
	}
	dput(subd);

	supermount_go_inactive(dentry->d_sb);
ret:
	return rc;
}

static int dir_dentry_compare(struct dentry *dentry, struct qstr *name1,
	struct qstr *name2)
{
	struct dentry *subd;

	supermount_debug("compare dir dentry %s, %s, %s\n", dentry->d_name.name,
		name1->name, name2->name);
	if ((subd = get_subfs_dentry(dentry)) != 0 && subd->d_op &&
			subd->d_op->d_compare)
	{
		int ret;
		supermount_debug("lowlevel\n");
		ret = subd->d_op->d_compare(subd, name1, name2);
		dput(subd);
		supermount_go_inactive(dentry->d_sb);
		return ret;
	}
	if (subd)
		dput(subd);
	supermount_go_inactive(dentry->d_sb);

	if (name1->len != name2->len)
		return 1;
	return memcmp(name1->name, name2->name, name1->len);
}

struct dentry_operations supermount_dir_dops = {
	dir_dentry_revalidate,
	NULL, //!dir_dentry_hash,
	NULL, //!dir_dentry_compare,
	NULL,
	NULL,
	NULL
};
