/*
 *  linux/fs/supermount/file.c
 *
 *  Original version:
 *      Copyright (C) 1995
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *  Rewriten for kernel 2.2 & 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@abc.cap.ru)
 *
 * $Id: file.c,v 1.8 2000/09/04 16:06:33 root Exp $
 */

#include <asm/segment.h>
#include <asm/current.h>

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/supermount_fs.h>
#include <linux/stat.h>
#include <linux/sched.h>
#include <linux/file.h>
#include <linux/poll.h>
#include <asm/smplock.h>

#define DEFAULT_POLLMASK (POLLIN | POLLOUT | POLLRDNORM | POLLWRNORM)

struct dentry *get_subfs_dentry(struct dentry *dentry);

static int file_revalidate_dentry(struct dentry *dentry, int flags)
{
	struct dentry *subd;
	int rc = 1;

	supermount_debug("file revalidate dentry %s\n", dentry->d_name.name);

	if (!supermount_revalidate(dentry->d_sb))
	{
		d_drop(dentry);
		return 0;
	}

	if ((subd = get_subfs_dentry(dentry)) != 0 && 
			subd->d_op && subd->d_op->d_revalidate)
	{
		supermount_debug("lowlevel\n");
		rc = subd->d_op->d_revalidate(subd, flags);
	}
	if (subd)
		dput(subd);
	else
	{
		rc = 0;
		d_drop(dentry);
	}

	supermount_go_inactive(dentry->d_sb);
	return rc;
}

struct dentry_operations supermount_file_dops = {
	file_revalidate_dentry,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

int supermount_readpage(struct file *file, struct page *page)
{
	struct file *subfile = (struct file *) file->private_data;
	struct inode *subinode;

	if (!subfile)
		return -EINVAL;
	subinode = subfile->f_dentry->d_inode;
	if (!subinode)
		return -EINVAL;

	supermount_debug("readpage %s\n", 
		file->f_dentry->d_name.name);

	supermount_debug("kl: %d\n", kernel_locked());

	if (subinode->i_mapping && subinode->i_mapping->a_ops &&
			subinode->i_mapping->a_ops->readpage)
	{
		int rc = subinode->i_mapping->a_ops->readpage(subfile, page);
		supermount_debug("kl: %d\n", kernel_locked());
		return rc;
	}
	return -ENOEXEC;
}

int supermount_writepage(struct page *page)
{
	if (page->mapping && page->mapping->host &&
			((struct inode*)page->mapping->host)->i_mapping &&
			((struct inode*)page->mapping->host)->i_mapping->a_ops &&
			((struct inode*)page->mapping->host)->i_mapping->a_ops->writepage)
		return ((struct inode*)page->mapping->host)->i_mapping->a_ops->writepage(page);
	return -EINVAL;
}

int supermount_prepare_write(struct file *file, struct page *page, unsigned offset, unsigned to)
{
	struct file *subfile = (struct file *) file->private_data;
	struct inode *subinode;

	if (!subfile)
		return -EINVAL;
	subinode = subfile->f_dentry->d_inode;
	if (!subinode)
		return -EINVAL;
	if (subinode->i_mapping && subinode->i_mapping->a_ops &&
			subinode->i_mapping->a_ops->prepare_write)
		return subinode->i_mapping->a_ops->prepare_write(subfile, page, offset, to);
	return -EINVAL;
}

int supermount_commit_write(struct file *file, struct page *page, unsigned offset, unsigned to)
{
	struct file *subfile = (struct file *) file->private_data;
	struct inode *subinode;

	if (!subfile)
		return -EINVAL;
	subinode = subfile->f_dentry->d_inode;
	if (!subinode)
		return -EINVAL;
	if (subinode->i_mapping && subinode->i_mapping->a_ops &&
			subinode->i_mapping->a_ops->commit_write)
		return subinode->i_mapping->a_ops->commit_write(subfile, page, offset, to);
	return -EINVAL;
}

struct address_space_operations supermount_file_aops = {
	readpage:	supermount_readpage,
	writepage:	supermount_writepage,
	prepare_write:	supermount_prepare_write,
	commit_write:	supermount_commit_write,
};
