/*
 *  linux/fs/supermount/inode.c
 *
 *  Original version:
 *      Copyright (C) 1995, 1997
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/inode.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/inode.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *      and           1993  Stephen Tweedie
 *
 *  Rewriten for kernel 2.2 & 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@abc.cap.ru)
 *
 * $Id: inode.c,v 1.30 2000/10/31 16:17:18 root Exp $
 */
#include <asm/segment.h>
#include <asm/system.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/supermount_fs.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/mm.h>
#include <asm/smplock.h>

/* Supermount open/close inode.  These functions maintain the
   superblock reference counts of active inodes; the subfs is
   suspended when that count reaches zero.

   Files are opened on any significant reference, and are not closed
   until they become fully dereferenced (during the last iput).
   Directories, on the other hand, are always closed unless they are
   active and opened; a directory referenced as CWD is not open.
*/

void supermount_iopen(struct super_block * sb)
{
	switch (sb->u.supermount_sb.s_state) {
	case SUPERMOUNT_UNMOUNTED:
		supermount_panic (sb, "supermount_iopen", 
				  "opening inode on unmounted subfs");
	case SUPERMOUNT_SUSPENDED:
		supermount_debug("locking door.\n");
		supermount_lock_door(sb, 1);
		break;
	default: 
	}
	
	if (!subfs_is_active(sb)) {
		supermount_debug ("going online.\n");
		sb->u.supermount_sb.s_state = 
			SUPERMOUNT_ONLINE;
	}
	atomic_inc(&sb->u.supermount_sb.s_opencount);
	supermount_debug("Opencount: %d\n", 
		atomic_read(&sb->u.supermount_sb.s_opencount));
}

void supermount_iclose(struct super_block * sb)
{
	atomic_dec(&sb->u.supermount_sb.s_opencount);
	supermount_debug("Opencount: %d\n", 
		atomic_read(&sb->u.supermount_sb.s_opencount));
	if (!subfs_is_active(sb))
		supermount_go_inactive(sb);
}

void supermount_read_inode (struct inode * inode)
{
	supermount_dump_inode(inode);
	
	inode->u.supermount_i.i_sb_version = 
		inode->i_sb->u.supermount_sb.s_version;
	
	switch (inode->i_ino) {
	case SUPERMOUNT_ROOT_INO:
		if (inode->i_sb->u.supermount_sb.s_state !=
		    SUPERMOUNT_UNMOUNTED) {
			supermount_panic (inode->i_sb, "supermount_read_inode",
					  "Help - trying to read root while "
					  "already mounted!");
		}
		/* Fall through */
	case SUPERMOUNT_HIDDEN_INO:
		inode->i_mode = inode->i_sb->u.supermount_sb.s_default_mode;
		inode->i_uid = 0;
		inode->i_gid = 0;
		inode->i_nlink = 1;
		inode->i_size = 0;
		inode->i_atime = CURRENT_TIME;
		inode->i_ctime = CURRENT_TIME;
		inode->i_mtime = CURRENT_TIME;
		inode->i_blksize = inode->i_sb->s_blocksize;
		inode->i_blocks = 0;
		inode->i_version = ++event;
		inode->i_op = (inode->i_ino == SUPERMOUNT_ROOT_INO)
			? &supermount_dir_iops : NULL;
		inode->i_fop = (inode->i_ino == SUPERMOUNT_ROOT_INO)
			? &supermount_dir_operations : NULL;
		return;
	default:
		;
	}
}

struct dentry *get_subfs_dentry_by_inode(struct inode *inode);

void supermount_write_inode (struct inode * inode, int sync)
{
	struct super_block *tmp = inode->i_sb->u.supermount_sb.s_subfs;
	struct dentry *dent;
	struct inode *subi;
	int rc = 0;

	dent = get_subfs_dentry_by_inode(inode);
	if (dent == NULL)
		goto ret;
	if ((subi = dent->d_inode) == NULL)
		goto dput_ret;
	
	supermount_dump_inode(inode);
	supermount_debug("kl: %d\n", kernel_locked());

	if (subi && tmp && tmp->s_op && tmp->s_op->write_inode)
	{
		subi->i_mode = inode->i_mode;
		subi->i_uid = inode->i_uid;
		subi->i_gid = inode->i_gid;
		subi->i_size = inode->i_size;
		subi->i_atime = inode->i_atime;
		subi->i_mtime = inode->i_mtime;
		subi->i_ctime = inode->i_ctime;
		rc = supermount_get_write_access(inode->i_sb);
		if (!rc)
		{
			subi->i_state |= I_LOCK;
			tmp->s_op->write_inode(subi, sync);
			subi->i_state &= ~I_LOCK;
			mark_subfs_dirty(inode->i_sb);
		}
		supermount_put_write_access(inode->i_sb);
	}

dput_ret:
	dput(dent);
ret:
	supermount_debug("kl: %d\n", kernel_locked());
	supermount_go_inactive(inode->i_sb);
	supermount_debug("kl: %d\n", kernel_locked());
}

int supermount_permission (struct inode * inode, int mask)
{
	struct inode *subi;
	struct dentry *subdent;
	int rc = -EIO;
	
	supermount_dump_inode(inode);
	
	if ((subdent = get_subfs_dentry_by_inode(inode)) == NULL)
		goto ret;
	if ((subi = subdent->d_inode) == NULL)
		goto dput_ret;
	
	rc = permission(subi, mask);
	supermount_debug("permission: %d, %d, %08x\n", rc, 
		(subi ? (int)subi->i_shadow : -1), inode->i_sb->s_flags);

dput_ret:
	dput(subdent);
ret:
	supermount_go_inactive(inode->i_sb);
	if (subdent != NULL)
		supermount_debug("inode: %d, count: %d\n", inode->i_ino, atomic_read(&subdent->d_count));
	return rc;
}
