/*
 *  linux/fs/supermount/namei.c
 *
 *  Original version:
 *      Copyright (C) 1995
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/namei.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/namei.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *
 *  Rewriten for kernel 2.2 & 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@abc.cap.ru)
 *
 * $Id: namei.c,v 1.34 2000/09/04 16:06:33 root Exp $
 */

/* This file handles almost all of the normal filesystem running of
   supermount.  We don't have to deal to much with the subfs
   interface, since we just pass subinodes out to the application on
   demand. */

#include <asm/segment.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/supermount_fs.h>
#include <linux/fcntl.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>

/*
 * Find a new superinode, avoiding collisions with old
 * obsolete superinodes. 
 */
struct inode *get_possible_super_inode(struct super_block *sb, ino_t ino)
{
	unsigned long new_ino, incr=1237;
	struct inode *superi;

	new_ino = ino + incr;
	for ( ; (superi = iget(sb, new_ino)) && 
			atomic_read(&superi->i_count) > 1 &&
			(inode_is_obsolete(superi) ||
			superi->u.supermount_i.i_subino != ino); ) 
	{
		iput(superi);
		new_ino += incr;
		new_ino &= 0xffffff;
		while (new_ino < 2)
		{
			new_ino += incr;
			incr = (incr << 1) ^ incr;
		}
		incr = (incr << 1) ^ incr;
	}

	supermount_dump_inode(superi);

	return superi;
}

extern struct file_operations supermount_file_operations;
extern struct address_space_operations supermount_file_aops;
extern struct inode_operations supermount_file_iops;

void supermount_update_super_inode(struct inode *superi, struct inode *subi)
{
	superi->i_mode = subi->i_mode;
	superi->i_uid = subi->i_uid;
	superi->i_gid = subi->i_gid;
	superi->i_nlink = subi->i_nlink;
	superi->i_size = subi->i_size;
	superi->i_atime = subi->i_atime;
	superi->i_ctime = subi->i_ctime;
	superi->i_mtime = subi->i_mtime;
	superi->i_blksize = subi->i_blksize;
	superi->i_blocks = subi->i_blocks;
	superi->i_rdev = subi->i_rdev;
	superi->i_version = ++event;

	if (S_ISDIR(superi->i_mode))
	{
		superi->i_op = &supermount_dir_iops;
		superi->i_fop = &supermount_dir_operations;
	}
	else
	{
		superi->i_op = &supermount_file_iops;
		superi->i_fop = &supermount_file_operations;
		superi->i_mapping->a_ops = &supermount_file_aops;
		superi->i_mapping->host = (void*) subi;
	}
	subi->i_shadow = superi->i_sb;
	subi->i_shadow_op = &supermount_shadow_iops;
}

#if 0
struct inode * supermount_create_super_inode(struct super_block *sb, struct inode *subi)
{
	struct inode *superi = 0;

	/*
	 * Find a new superinode, avoiding collisions with old
	 * obsolete superinodes.
	 */

	superi = get_possible_super_inode(sb, subi->i_ino);

	superi->u.supermount_i.i_subino = subi->i_ino;

	/*
	 * If we found a previously obsolete inode which is now no
	 * longer used, we can safely remove the obsolesence flag. 
	 */
	if (inode_is_obsolete(superi)) {
		superi->u.supermount_i.i_sb_version = 
			superi->i_sb->u.supermount_sb.s_version;
	}

	supermount_update_super_inode(superi, subi);

	return superi;
}
#endif

/*
 * Try to find superinode for subinode and make superinode mirror
 * subinode.
 */

struct inode * supermount_attach(struct super_block *sb, struct dentry *subd) 
{
	struct inode *subi = subd->d_inode;
	struct inode *superi = 0; /* to shut gcc up */
	
	supermount_debug ("inode = %ld\n", subi->i_ino);

	/*
	 * Find a new superinode, avoiding collisions with old
	 * obsolete superinodes.
	 */

	superi = get_possible_super_inode(sb, subi->i_ino);

	/*
	 * We now have a new superinode which has been refcounted by
	 * one.  Next we have to attach the subinode and subdentry to it. 
	 */

	superi->u.supermount_i.i_subino = subi->i_ino;

	/*
	 * If we found a previously obsolete inode which is now no
	 * longer used, we can safely remove the obsolesence flag. 
	 */
	if (inode_is_obsolete(superi)) {
		superi->u.supermount_i.i_sb_version = 
			superi->i_sb->u.supermount_sb.s_version;
	}

	supermount_update_super_inode(superi, subi);

	return superi;
}

/*
 * Attach superinode to subinode and instantiate dentry with this inode
 */

static void attach_and_prepare(struct super_block *sb, struct dentry *dentry,
							   struct dentry *subd, int create)
{
	struct inode *superi, *subi = subd->d_inode;

	if (dentry->d_inode != NULL)
	{
		supermount_update_super_inode(dentry->d_inode, subi);
		return;
	}

	superi = supermount_attach(sb, subd);

	if (create)
		d_instantiate(dentry, superi);
	else
		d_add(dentry, superi);

	if (S_ISDIR(subi->i_mode))
		dentry->d_op = &supermount_dir_dops;
	else
		dentry->d_op = &supermount_file_dops;

	supermount_dump_inode(superi);
	supermount_dump_normal_inode(subi);
	supermount_debug("dentry count: %d/%d\n", atomic_read(&dentry->d_count), (subd ? atomic_read(&subd->d_count) : -1));
}

struct dentry *get_subfs_dentry(struct dentry *dentry);
struct dentry *get_subfs_dentry_by_inode(struct inode *inode);

void print_tree(struct dentry *root);

/**** kernel locked, dir->i_sem held ****/

struct dentry *supermount_lookup (struct inode *dir, struct dentry *dentry)
{
//	struct inode *subdir;
	struct dentry *err, *subdent = NULL, *subfs_dir;

	supermount_debug("lookup: %s (parent: %08lx), directory: %ld\n",
		dentry->d_name.name, (unsigned long)dentry->d_parent,
		dir->i_ino);

	subfs_dir = get_subfs_dentry_by_inode(dir);
	err = ERR_PTR(-ENOENT);
	if (subfs_dir == NULL)
		goto ret;
	if (subfs_dir->d_inode == NULL)
		goto dirput_ret;

	subdent = d_alloc(subfs_dir, &dentry->d_name);
	if (!subdent)
	{
		err = ERR_PTR(-ENOMEM);
		goto dirput_ret;
	}

	supermount_debug("lowlevel lookup: %s\n", subdent->d_name.name);

	err = subfs_dir->d_inode->i_op->lookup(subfs_dir->d_inode, subdent);

	if (err)
	{
		supermount_debug("!!!!! WARNING !!!!! lowlevel error: %ld\n", (long)err);
		goto dput_ret;
	}

	if (!subdent->d_inode)
	{
		/* not found */
		d_add(dentry, NULL);
		goto dput_ret;
	}

//	supermount_dump_normal_inode(subdir);
//	supermount_dump_normal_inode(subdent->d_inode);

	/* It worked, so now create a shadow supermount inode for the
           result... */
	attach_and_prepare(dir->i_sb, dentry, subdent, 0);
dput_ret:
	dput(subdent);
dirput_ret:
	dput(subfs_dir);
ret:
	supermount_debug("%d\n", atomic_read(&subdent->d_count));
	supermount_go_inactive(dir->i_sb);
	return err;
}

/**** kernel locked, dir->i_zombie held, dir->i_sem held ****/

int supermount_create (struct inode * dir, struct dentry *dentry, int mode)
{
//	struct inode *subdir;
	struct dentry *subdent = NULL, *subfs_dir;
	int rc = -ENOENT;

	supermount_debug("create: %s (parent: %08lx), directory: %ld\n",
		dentry->d_name.name, (unsigned long)dentry->d_parent,
		dir->i_ino);

	subfs_dir = get_subfs_dentry_by_inode(dir);
	if (subfs_dir == NULL)
		goto ret;

	if (subfs_dir->d_inode == NULL)
		goto dirput_ret;

	if (!subfs_dir->d_inode->i_op || !subfs_dir->d_inode->i_op->create) {
		rc = -EPERM;
		goto dirput_ret;
	}

	supermount_debug("dentry hashed: %d\n", !list_empty(&dentry->d_hash));

	if ((subdent = get_subfs_dentry(dentry)) == NULL)
	{
		subdent = d_alloc(subfs_dir, &dentry->d_name);
		if (!subdent)
		{
			rc = -ENOMEM;
			goto dirput_ret;
		}
	}

	if (subdent->d_inode && !dentry->d_inode)
		goto attach;

	rc = supermount_get_write_access(dir->i_sb);
	if (rc)
		goto dput_ret;

	supermount_debug("lowlevel create: %s\n", subdent->d_name.name);
	rc = subfs_dir->d_inode->i_op->create(subfs_dir->d_inode, subdent, mode);

	supermount_debug("mode: %03o\n", subdent->d_inode->i_mode);
	mark_subfs_dirty(dir->i_sb);
	supermount_put_write_access(dir->i_sb);

	if (rc)
	{
		supermount_debug("!!!!! WARNING !!!!! lowlevel error: %d\n", rc);
		goto dput_ret;
	}

	if (!subdent->d_inode)
	{
		supermount_debug("!!!!! WARNING !!!!! lowlevel didn't return inode\n");
		d_instantiate(dentry, NULL);
		goto dput_ret;
	}

	supermount_dump_normal_inode(subfs_dir->d_inode);
	supermount_dump_normal_inode(subdent->d_inode);

	/* It worked, so now create a shadow supermount inode for the
           result... */
attach:
	attach_and_prepare(dir->i_sb, dentry, subdent, 1);

dput_ret:
	dput(subdent);
dirput_ret:
	dput(subfs_dir);
ret:
	supermount_debug("%d\n", atomic_read(&subdent->d_count));
	supermount_go_inactive(dir->i_sb);
	return rc;
}

inline void update_nlink(struct inode *superi, struct inode *subi)
{
	superi->i_nlink = subi->i_nlink;
}

inline void update_ctime(struct inode *superi, struct inode *subi)
{
	superi->i_ctime = subi->i_ctime;
}

inline void update_mtime(struct inode *superi, struct inode *subi)
{
	superi->i_mtime = subi->i_mtime;
}

/**** kernel locked, dir->i_zombie held, dir->i_sem held ****/

int supermount_mkdir (struct inode * dir, struct dentry *dentry, int mode)
{
//	struct inode *subdir;
	struct dentry *subdent, *subfs_dir;
	int rc = -ENOENT;

	supermount_debug("mkdir: %s (parent: %08lx), directory: %ld\n",
		dentry->d_name.name, (unsigned long)dentry->d_parent,
		dir->i_ino);

	subfs_dir = get_subfs_dentry_by_inode(dir);
	if (subfs_dir == NULL)
		goto ret;

	if (subfs_dir->d_inode == NULL)
		goto dirput_ret;

	if (!subfs_dir->d_inode->i_op || !subfs_dir->d_inode->i_op->create) {
		rc = -EPERM;
		goto dirput_ret;
	}

	if ((subdent = get_subfs_dentry(dentry)) == NULL)
	{
		subdent = d_alloc(subfs_dir, &dentry->d_name);
		if (!subdent)
		{
			rc = -ENOMEM;
			goto dirput_ret;
		}
	}

	if (subdent->d_inode && !dentry->d_inode)
		goto attach;

	rc = supermount_get_write_access(dir->i_sb);
	if (rc) 
		goto dput_ret;

	supermount_debug("lowlevel mkdir: %s\n", subdent->d_name.name);
	rc = subfs_dir->d_inode->i_op->mkdir(subfs_dir->d_inode, subdent, mode);
	supermount_debug("mode: %03o\n", subdent->d_inode->i_mode);

	update_nlink(dir, subfs_dir->d_inode);

	mark_subfs_dirty(dir->i_sb);
	supermount_put_write_access(dir->i_sb);
	supermount_go_inactive(dir->i_sb);

	if (rc)
	{
		supermount_debug("!!!!! WARNING !!!!! lowlevel error: %d\n", rc);
		goto dput_ret;
	}

	if (!subdent->d_inode)
	{
		supermount_debug("!!!!! WARNING !!!!! lowlevel didn't return inode\n");
		d_instantiate(dentry, NULL);
		goto dput_ret;
	}

	supermount_dump_normal_inode(subfs_dir->d_inode);
	supermount_dump_normal_inode(subdent->d_inode);

	/* It worked, so now create a shadow supermount inode for the
           result... */
attach:
	attach_and_prepare(dir->i_sb, dentry, subdent, 1);
dput_ret:
	dput(subdent);
dirput_ret:
	dput(subfs_dir); /* We don't need it anymore*/
ret:
	supermount_go_inactive(dir->i_sb);	
	return rc;
}

/**** kernel locked, dir->i_zombie held, dir->i_sem held */

int supermount_unlink (struct inode *dir, struct dentry *dentry)
{
//	struct inode *subdir;
	struct dentry *subfs_dir, *subdent;
	int rc = -ENOENT;

	supermount_debug("unlink: %s (parent: %08lx), directory: %ld\n",
		dentry->d_name.name, (unsigned long)dentry->d_parent,
		dir->i_ino);

	subfs_dir = get_subfs_dentry_by_inode(dir);
	if (subfs_dir == NULL)
		goto ret;

	if (subfs_dir->d_inode == NULL)
		goto dirput_ret;

	if (!subfs_dir->d_inode->i_op || !subfs_dir->d_inode->i_op->unlink) {
		rc = -EPERM;
		goto dirput_ret;
	}

	if ((subdent = get_subfs_dentry(dentry)) == NULL)
		goto dirput_ret;

	rc = supermount_get_write_access(dir->i_sb);
	if (rc) 
		goto dput_ret;

	supermount_debug("lowlevel unlink: %s\n", dentry->d_name.name);

	rc = subfs_dir->d_inode->i_op->unlink(subfs_dir->d_inode, subdent);

	update_nlink(dir, subfs_dir->d_inode);
	update_nlink(dentry->d_inode, subdent->d_inode);
	update_ctime(dir, subfs_dir->d_inode);
	update_mtime(dir, subfs_dir->d_inode);
	update_ctime(dentry->d_inode, subdent->d_inode);

	if (!rc)
		d_delete(subdent);

	/* If subdent->d_count == 1 (which means that this function is the
	 * only holder of dentry then d_delete called a little earlier will
	 * dentry_iput subdent, so if its inode has no other links inode_delete
	 * will be called at this point, otherwise inode_delete will be
	 * called when other link(s) to this inode is deleted. So we
	 * can safely put_write_access now.
	 */
	/* Then, if subdent->d_count != 1 then last iput will be called on last 
	 * dput (if inode has no other links) so we will have to check for 
	 * ->i_nlink == 0 in supermount_release
	 */

	mark_subfs_dirty(dir->i_sb);
	supermount_put_write_access(dir->i_sb);

dput_ret:
	dput(subdent);
dirput_ret:
	dput(subfs_dir);
ret:
	supermount_go_inactive(dir->i_sb);
	return rc;
}

/**** kernel locked, dir->i_zombie held, dentry->d_inode->i_zombie held ****/
/**** dir->i_sem held ****/

int supermount_rmdir (struct inode *dir, struct dentry *dentry)
{
//	struct inode *subdir;
	struct dentry *subdent, *subfs_dir;
	int rc = -ENOENT;

	supermount_debug("rmdir: %s (parent: %08lx), directory: %ld\n",
		dentry->d_name.name, (unsigned long)dentry->d_parent,
		dir->i_ino);

	subfs_dir = get_subfs_dentry_by_inode(dir);
	if (subfs_dir == NULL)
		goto ret;

	if (subfs_dir->d_inode == NULL)
		goto dirput_ret;

	if (!subfs_dir->d_inode->i_op || !subfs_dir->d_inode->i_op->rmdir) {
		rc = -EPERM;
		goto dirput_ret;
	}

	if ((subdent = get_subfs_dentry(dentry)) == NULL)
		goto dirput_ret;

	rc = supermount_get_write_access(dir->i_sb);
	if (rc) 
		goto dput_ret;

	d_unhash(subdent); /* it makes dget on subdent */
	supermount_debug("lowlevel rmdir: %s\n", subdent->d_name.name);
	rc = subfs_dir->d_inode->i_op->rmdir(subfs_dir->d_inode, subdent);
	dput(subdent); /* dget was in d_unhash */

	update_nlink(dir, subfs_dir->d_inode);
	update_nlink(dentry->d_inode, subdent->d_inode);
	update_ctime(dir, subfs_dir->d_inode);
	update_mtime(dir, subfs_dir->d_inode);
	update_ctime(dentry->d_inode, subdent->d_inode);

	/* See comment in supermount_unlink */
	if (!rc)
		d_delete(subdent);       /* This also frees the inode */

	mark_subfs_dirty(dir->i_sb);
	supermount_put_write_access(dir->i_sb);

dput_ret:
	dput(subdent);
dirput_ret:
	dput(subfs_dir);
ret:
	supermount_go_inactive(dir->i_sb);
	return rc;
}

/**** kernel locked, dir->i_zombie held, dir->i_sem held ****/

int supermount_mknod (struct inode *dir, struct dentry *dentry, int mode, int dev)
{
//	struct inode *subdir;
	struct dentry *subdent, *subfs_dir;
	int rc = -ENOENT;

	supermount_debug("mknod: %s (parent: %08lx), directory: %ld\n",
		dentry->d_name.name, (unsigned long)dentry->d_parent,
		dir->i_ino);

	subfs_dir = get_subfs_dentry_by_inode(dir);
	if (subfs_dir == NULL)
		goto ret;

	if (subfs_dir->d_inode == NULL)
		goto dirput_ret;

	if (!subfs_dir->d_inode->i_op || !subfs_dir->d_inode->i_op->mknod) {
		rc = -EPERM;
		goto dirput_ret;
	}

	if ((subdent = get_subfs_dentry(dentry)) == NULL)
	{
		subdent = d_alloc(subfs_dir, &dentry->d_name);
		if (!subdent)
		{
			rc = -ENOMEM;
			goto dirput_ret;
		}
	}

	if (subdent->d_inode && !dentry->d_inode)
		goto attach;

	rc = supermount_get_write_access(dir->i_sb);
	if (rc) 
		goto dput_ret;

	supermount_debug("lowlevel mknod: %s\n", subdent->d_name.name);
	rc = subfs_dir->d_inode->i_op->mknod(subfs_dir->d_inode, subdent, mode, dev);
	/* Now we have subinode, attached to subdent. */

	supermount_debug("mode: %03o\n", subdent->d_inode->i_mode);
	mark_subfs_dirty(dir->i_sb);
	supermount_put_write_access(dir->i_sb);

	if (rc)
	{
		supermount_debug("!!!!! WARNING !!!!! lowlevel error: %d\n", rc);
		goto dput_ret;
	}

	if (!subdent->d_inode)
	{
		supermount_debug("!!!!! WARNING !!!!! lowlevel didn't return inode\n");
		d_instantiate(dentry, NULL);
		goto dput_ret;
	}

	supermount_dump_normal_inode(subfs_dir->d_inode);
	supermount_dump_normal_inode(subdent->d_inode);

	/* It worked, so now create a shadow supermount inode for the
           result... */
attach:
	attach_and_prepare(dir->i_sb, dentry, subdent, 1);
dput_ret:
	dput(subdent);
dirput_ret:
	dput(subfs_dir);
ret:
	supermount_go_inactive(dir->i_sb);
	return rc;
}

/**** kernel locked, dir->i_zombie held, dir->i_sem held ****/

int supermount_symlink (struct inode * dir, struct dentry *dentry,
	const char * symname)
{
//	struct inode *subdir;
	struct dentry *subdent, *subfs_dir;
	int rc = -ENOENT;

	supermount_debug("symlink: %s (parent: %08lx), directory: %ld\n",
		dentry->d_name.name, (unsigned long)dentry->d_parent,
		dir->i_ino);

	subfs_dir = get_subfs_dentry_by_inode(dir);
	if (subfs_dir == NULL)
		goto ret;

	if (subfs_dir->d_inode == NULL)
		goto dirput_ret;

	if (!subfs_dir->d_inode->i_op || !subfs_dir->d_inode->i_op->symlink) {
		rc = -EPERM;
		goto dirput_ret;
	}

	if ((subdent = get_subfs_dentry(dentry)) == NULL)
	{
		subdent = d_alloc(subfs_dir, &dentry->d_name);
		if (!subdent)
		{
			rc = -ENOMEM;
			goto dirput_ret;
		}
	}

	if (subdent->d_inode && !dentry->d_inode)
		goto attach;

	rc = supermount_get_write_access(dir->i_sb);
	if (rc) 
		goto dput_ret;

	supermount_debug("lowlevel symlink: %s->%s\n", symname, subdent->d_name.name);
	rc = subfs_dir->d_inode->i_op->symlink(subfs_dir->d_inode, subdent, symname);
	/* Now we have subinode, attached to subdent. */

	supermount_debug("mode: %03o\n", subdent->d_inode->i_mode);
	mark_subfs_dirty(dir->i_sb);
	supermount_put_write_access(dir->i_sb);

	if (rc)
	{
		supermount_debug("!!!!! WARNING !!!!! lowlevel error: %d\n", rc);
		goto dput_ret;
	}

	if (!subdent->d_inode)
	{
		supermount_debug("!!!!! WARNING !!!!! lowlevel didn't return inode\n");
		d_instantiate(dentry, NULL);
		goto dput_ret;
	}

	supermount_dump_normal_inode(subfs_dir->d_inode);
	supermount_dump_normal_inode(subdent->d_inode);

	/* It worked, so now create a shadow supermount inode for the
           result... */
attach:
	attach_and_prepare(dir->i_sb, dentry, subdent, 1);
dput_ret:	
	dput(subdent);
dirput_ret:
	dput(subfs_dir);
ret:
	supermount_go_inactive(dir->i_sb);
	return rc;
}

/**** kernel locked, olddir->i_zombie held, newdir->i_zombie held, ****/
/**** olddir->i_sem held, newdir->i_sem held,                      ****/
/**** olddir->i_sb->s_vfs_rename_sem held                          ****/
/**** if newdentry is directory and exist newdentry->i_zombie held ****/

int supermount_rename (struct inode * olddir, struct dentry *olddentry,
						struct inode * newdir, struct dentry *newdentry)
{
//	struct inode *oldsubdir, *newsubdir;
	struct dentry *oldsubdent, *newsubdent, *oldsubfs_dir, *newsubfs_dir;
	int rc = -ENOENT;

	supermount_debug("rename: %s(%08x)->%s(%08x) (parent: %08lx/%08lx), directory: %ld/%ld\n",
		olddentry->d_name.name, (unsigned int)olddentry->d_inode,
		newdentry->d_name.name, (unsigned int)newdentry->d_inode,
		(unsigned long)olddentry->d_parent, (unsigned long)newdentry->d_parent, 
		olddir->i_ino, newdir->i_ino);

	oldsubdent = get_subfs_dentry(olddentry);
	if (!oldsubdent)
		goto ret;

	oldsubfs_dir = get_subfs_dentry_by_inode(olddir);
	if (!oldsubfs_dir)
		goto olddput_ret;

	if (oldsubfs_dir->d_inode == NULL)
		goto olddirput_ret;

	if (!oldsubfs_dir->d_inode->i_op || !oldsubfs_dir->d_inode->i_op->rename) {
		rc = -EPERM;
		goto olddirput_ret;
	}

	newsubfs_dir = get_subfs_dentry_by_inode(newdir);
	if (newsubfs_dir == NULL)
		goto olddirput_ret;

	newsubdent = get_subfs_dentry(newdentry);
	if (!newsubdent)
	{
		rc = -ENOMEM;
		goto newdirput_ret;
	}

	rc = supermount_get_write_access(olddir->i_sb);
	if (rc)
		goto newdput_ret;

	supermount_debug("lowlevel rename: %s->%s\n", oldsubdent->d_name.name,
												  newsubdent->d_name.name);
	rc = oldsubfs_dir->d_inode->i_op->rename(oldsubfs_dir->d_inode, oldsubdent, newsubfs_dir->d_inode, newsubdent);

	update_nlink(olddir, oldsubfs_dir->d_inode);
	update_nlink(newdir, newsubfs_dir->d_inode);
	update_nlink(olddentry->d_inode, oldsubdent->d_inode);
//	update_nlink(newdentry->d_inode, newsubdent->d_inode);
	update_ctime(olddir, oldsubfs_dir->d_inode);
	update_mtime(olddir, oldsubfs_dir->d_inode);
	update_ctime(olddentry->d_inode, oldsubdent->d_inode);
//	update_ctime(newdentry->d_inode, newsubdent->d_inode);

	supermount_put_write_access(olddir->i_sb);
newdput_ret:
	dput(newsubdent);
newdirput_ret:
	dput(newsubfs_dir);
olddirput_ret:
	dput(oldsubfs_dir);
olddput_ret:
	dput(oldsubdent);
ret:
	supermount_go_inactive(olddir->i_sb);
	return rc;
}
